@font-face{font-family:Source Code Pro;font-style:normal;font-weight:400;src:url(./SourceCodePro-Regular-BoGDH3YB.ttf) format("truetype")}@font-face{font-family:Source Code Pro;font-style:normal;font-weight:600;src:url(./SourceCodePro-SemiBold-R6YWw784.ttf) format("truetype")}@font-face{font-family:Source Code Pro;font-style:normal;font-weight:700;src:url(./SourceCodePro-Bold-C5Vtz9WP.ttf) format("truetype")}
